/* -*- buffer-read-only: t -*- vi: set ro:
   THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT.
*/
#line 973
#ifdef HAVE_CONFIG_H
#line 973
# include <config.h>
#line 973
#endif
#line 973
#include <sys/types.h>
#line 973

#line 973
#include "mailfromd.h"
#line 973
#include "prog.h"
#line 973
#include "builtin.h"
#line 973

#line 210 "db.bi"
static mu_debug_handle_t debug_handle;
#line 973 "db.bi"

#line 769 "../../src/builtin/snarf.m4"

#line 1 "db.bi"
/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define DEFAULT_DB_MODE 0640
#include <fnmatch.h>

struct db_prop {                /* Database properties */
	char *pat;              /* Database name pattern */
	mode_t mode;            /* File mode */
	int null;               /* Null byte */
	mu_url_t hint;          /* Hint to use instead of the default one */
};

static mu_list_t db_prop_list;

static int
db_prop_compare(const void *item, const void *data)
{
	const struct db_prop *prop = item;
	const char *name = data;
	return strcmp(prop->pat, name);
}

static int
strtomode(char *str, mode_t *pmode)
{
	mode_t mode = 0;
	struct { char c; unsigned mask; } modetab[] = {
		{ 'r', S_IRUSR },
		{ 'w', S_IWUSR },
		{ 'x', S_IXUSR },
		{ 'r', S_IRGRP },
		{ 'w', S_IWGRP },
		{ 'x', S_IXGRP },
		{ 'r', S_IROTH },
		{ 'w', S_IWOTH },
		{ 'x', S_IXOTH },
		{ 0 }
	};
	int i;

	for (i = 0; modetab[i].c; i++) {
		if (!str[i])
			return i + 1;
		if (str[i] == modetab[i].c)
			mode |= modetab[i].mask;
		else if (str[i] != '-')
			return i + 1;
	}
	if (str[i])
		return i + 1;
	*pmode = mode;
	return 0;
}

static int
is_url(const char *p)
{
	for (; *p && c_isalnum(*p); p++)
		;
	return *p == ':';
}
		
/* #pragma dbprop <name> [null] [mode] [hint]
   At least one of the bracketed parameters must be present.  Two or more
   parameters can be given in arbitrary order.
 */
#line 81 "db.bi"

#line 81
static void _pragma_dbprop (int argc, char **argv, const char *text)
#line 81

{
	int null = 0;
	mode_t mode = DEFAULT_DB_MODE;
	struct db_prop *prop;
	int rc;
	char *pat;
	mu_url_t hint = NULL;

	--argc;
	pat = *++argv;
	
	while (--argc) {
		char *p = *++argv;

		if (strcmp(p, "null") == 0)
			null = 1;
		else if (c_isdigit(*p)) {
			unsigned long n = strtoul(p, &p, 8);
			if (*p || (mode = n) != n) {
				parse_error(_("bad numeric file mode"));
				return;
			}
		} else if (is_url(p)) {
			rc = mu_url_create(&hint, p);
			if (rc) {
				parse_error(_("not a valid URL: %s"),
					    mu_strerror(rc));
				return;
			}
		} else if (rc = strtomode(p, &mode)) {
			parse_error(_("bad symbolic file mode (near %s)"),
				    p + rc - 1);
			return;
		}
	}

	if (!db_prop_list) {
		rc = mu_list_create(&db_prop_list);
		if (rc) {
			parse_error(_("cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
		mu_list_set_comparator(db_prop_list, db_prop_compare);
	}

	if (mu_list_locate(db_prop_list, pat, (void**) &prop)) {
		prop = xmalloc(sizeof(*prop));
		prop->pat = xstrdup(pat);
		rc = mu_list_append(db_prop_list, prop);
		if (rc) {
			parse_error(_("Cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
	}
	prop->mode = mode;
	prop->null = null;
	prop->hint = hint;
}

const struct db_prop *
db_prop_lookup(const char *name)
{
	mu_iterator_t itr;
	const struct db_prop *found = NULL;
	
	if (db_prop_list
	    && mu_list_get_iterator(db_prop_list, &itr) == 0) {
		for (mu_iterator_first(itr);
		     !mu_iterator_is_done(itr);
		     mu_iterator_next(itr)) {
			const struct db_prop *p;
			mu_iterator_current(itr, (void**)&p);
			if (strcmp(p->pat, name) == 0) {
				found = p;
				break;
			} else if (fnmatch(p->pat, name, 0) == 0)
				found = p;
		}
		mu_iterator_destroy(&itr);
	}
	return found;	
}

int 
_open_dbm(mu_dbm_file_t *pdb, char *dbname, int access, int mode,
	  mu_url_t hint)
{
	mu_dbm_file_t db;
	int rc;
	mu_url_t url;

	if (!hint)
		hint = mu_dbm_get_hint();
	rc = mu_url_create_hint(&url, dbname, 0, hint);
	if (rc) {
		mu_error(_("cannot create database URL for %s: %s"),
			 dbname, mu_strerror(rc));
		return rc;
	}
	rc = mu_dbm_create_from_url(url, &db,
				    mf_file_mode_to_safety_criteria(mode));
	mu_url_destroy(&url);
	if (rc)
		return rc;

	rc = mu_dbm_safety_check(db);
	if (rc && rc != ENOENT) {
		mu_error(_("%s fails safety check: %s"),
			 dbname, mu_strerror(rc));
		mu_dbm_destroy(&db);
		return rc;
	}

	rc = mu_dbm_open(db, access, mode);
	if (rc) {
		mu_dbm_destroy(&db);
		return rc;
	}
	*pdb = db;
	return rc;
}


#define LOOKUP_NULL_BYTE 0x1
#define LOOKUP_TEST_ONLY 0x2


#line 210

#line 210
static int
#line 210
dbmap_lookup(eval_environ_t env, char *dbname, const char *keystr,
#line 210
	     const char *defval, const struct db_prop *prop, int flags)
#line 210
{
#line 210
	int rc;
#line 210
	mu_dbm_file_t db;
#line 210
	struct mu_dbm_datum key;
#line 210
	struct mu_dbm_datum contents;
#line 210
	
#line 210
	if (!defval)
#line 210
		defval = "";
#line 210
	rc = _open_dbm(&db, dbname, MU_STREAM_READ,
#line 210
		       prop ? prop->mode : DEFAULT_DB_MODE,
#line 210
		       prop ? prop->hint : NULL);
#line 210
	if (rc)
#line 210
		(
#line 210
	env_throw_bi(env, mfe_dbfailure, NULL, _("mf_dbm_open(%s) failed: %s"),dbname,mu_strerror(rc))
#line 210
);
#line 210

#line 210
	memset(&key, 0, sizeof key);
#line 210
	memset(&contents, 0, sizeof contents);
#line 210
	key.mu_dptr = (void*) keystr;
#line 210
	key.mu_dsize = strlen(keystr);
#line 210
	if (flags & LOOKUP_NULL_BYTE)
#line 210
		key.mu_dsize++;
#line 210
	rc = mu_dbm_fetch(db, &key, &contents);
#line 210
	if (rc && rc != MU_ERR_NOENT) {
#line 210
		mu_dbm_destroy(&db);
#line 210
		(
#line 210
	env_throw_bi(env, mfe_dbfailure, NULL, _("cannot fetch data: %s"),mu_dbm_strerror(db))
#line 210
);
#line 210
	}
#line 210
	
#line 210 "db.bi"

#line 210
mu_debug(debug_handle, MU_DEBUG_TRACE5,("Looking up %s: %s", keystr, rc ? "false" : "true"));
#line 210
	if (flags & LOOKUP_TEST_ONLY) 
#line 210
		push(env, (STKVAL) (rc == 0));
#line 210
	else {
#line 210
		if (rc) {
#line 210
			if (defval)
#line 210
				pushs(env, (STKVAL)defval);
#line 210
			else
#line 210
				push(env, 0);
#line 210
		} else if (((char*)contents.mu_dptr)[contents.mu_dsize-1]) {
#line 210
			size_t off;
#line 210
			size_t len = contents.mu_dsize;
#line 210
			char *s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
#line 210
			memcpy(s, contents.mu_dptr, len);
#line 210
			s[len] = 0;
#line 210
			push(env, (STKVAL) off);
#line 210
		} else
#line 210
			pushs(env, contents.mu_dptr);
#line 210
	}
#line 210
	mu_dbm_datum_free(&contents);
#line 210
	mu_dbm_destroy(&db);
#line 210
	return rc;
#line 210
}
#line 210

#line 210

#line 210

#line 268



void
#line 271
bi_dbmap(eval_environ_t env)
#line 271

#line 271

#line 271 "db.bi"
{
#line 271
	
#line 271

#line 271
long __bi_argcnt;
#line 271
char * MFL_DATASEG dbname;
#line 271
        char * MFL_DATASEG key;
#line 271
        long  null;
#line 271
        
#line 271
get_string_arg(env, 1, &dbname);
#line 271
        get_string_arg(env, 2, &key);
#line 271
        get_numeric_arg(env, 3, &null);
#line 271
        
#line 271
get_numeric_arg(env, 0, &__bi_argcnt);
#line 271
        adjust_stack(env, __bi_argcnt + 1);
#line 271

#line 271

#line 271
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 271
		prog_trace(env, "dbmap %s %s %lu",dbname, key, ((__bi_argcnt > 2) ? null : 0));;
#line 271

{
	const struct db_prop *prop = db_prop_lookup(dbname);
	
#line 274
dbmap_lookup(env,dbname,key,NULL,prop,LOOKUP_TEST_ONLY | 
		         (((__bi_argcnt > 2) ? null : prop && prop->null)
			   ? LOOKUP_NULL_BYTE : 0));
#line 279
}

#line 280
        env_function_cleanup_flush(env, NULL);
#line 280
	return;
#line 280
}

void
#line 282
bi_dbget(eval_environ_t env)
#line 282

#line 282

#line 282 "db.bi"
{
#line 282
	
#line 282

#line 282
long __bi_argcnt;
#line 282
char * MFL_DATASEG dbname;
#line 282
        char * MFL_DATASEG key;
#line 282
        char * MFL_DATASEG defval;
#line 282
        long  null;
#line 282
        
#line 282
get_string_arg(env, 1, &dbname);
#line 282
        get_string_arg(env, 2, &key);
#line 282
        get_string_arg(env, 3, &defval);
#line 282
        get_numeric_arg(env, 4, &null);
#line 282
        
#line 282
get_numeric_arg(env, 0, &__bi_argcnt);
#line 282
        adjust_stack(env, __bi_argcnt + 1);
#line 282

#line 282

#line 282
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 282
		prog_trace(env, "dbget %s %s %s %lu",dbname, key, ((__bi_argcnt > 2) ? defval : ""), ((__bi_argcnt > 3) ? null : 0));;

{
	const struct db_prop *prop = db_prop_lookup(dbname);
	
#line 286
dbmap_lookup(env,dbname,key,((__bi_argcnt > 2) ? defval : 0),prop,((__bi_argcnt > 3) ? null : prop && prop->null)
		      ? LOOKUP_NULL_BYTE : 0);
#line 291
}

#line 292
        env_function_cleanup_flush(env, NULL);
#line 292
	return;
#line 292
}

void
#line 294
bi_dbput(eval_environ_t env)
#line 294

#line 294

#line 294 "db.bi"
{
#line 294
	
#line 294

#line 294
long __bi_argcnt;
#line 294
char *  dbname;
#line 294
        char *  keystr;
#line 294
        char *  value;
#line 294
        long  null;
#line 294
        long  mode;
#line 294
        
#line 294
get_string_arg(env, 1, &dbname);
#line 294
        get_string_arg(env, 2, &keystr);
#line 294
        get_string_arg(env, 3, &value);
#line 294
        get_numeric_arg(env, 4, &null);
#line 294
        get_numeric_arg(env, 5, &mode);
#line 294
        
#line 294
get_numeric_arg(env, 0, &__bi_argcnt);
#line 294
        adjust_stack(env, __bi_argcnt + 1);
#line 294

#line 294

#line 294
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 294
		prog_trace(env, "dbput %s %s %s %lu %lu",dbname, keystr, value, ((__bi_argcnt > 3) ? null : 0), ((__bi_argcnt > 4) ? mode : 0));;

{
	int rc;
	mu_dbm_file_t db;
	struct mu_dbm_datum key;
	struct mu_dbm_datum contents;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	rc = _open_dbm(&db, dbname, MU_STREAM_RDWR,
		       ((__bi_argcnt > 4) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
#line 306
		       prop ? prop->hint : NULL);
	if (rc)
		(
#line 308
	env_throw_bi(env, mfe_dbfailure, "dbput", _("mf_dbm_open(%s) failed: %s"),dbname,mu_strerror(rc))
#line 308
);
#line 312
	memset(&key, 0, sizeof key);
	key.mu_dptr = keystr;
	key.mu_dsize = strlen(keystr);
	if (((__bi_argcnt > 3) ? null : prop && prop->null))
		key.mu_dsize++;

	memset(&contents, 0, sizeof contents);
	contents.mu_dptr = value;
	contents.mu_dsize = strlen(value) + 1;
	
	rc = mu_dbm_store(db, &key, &contents, 1);
	if (rc) {
		const char *errstr = mu_dbm_strerror(db);
		mu_dbm_destroy(&db);
		(
#line 326
	env_throw_bi(env, mfe_dbfailure, "dbput", _("failed to insert data to %s: %s %s: %s"),dbname,keystr,value,errstr)
#line 326
);
#line 332
	}
	mu_dbm_destroy(&db);
}

#line 335
        env_function_cleanup_flush(env, NULL);
#line 335
	return;
#line 335
}

void
#line 337
bi_dbinsert(eval_environ_t env)
#line 337

#line 337

#line 337 "db.bi"
{
#line 337
	
#line 337

#line 337
long __bi_argcnt;
#line 337
char *  dbname;
#line 337
        char *  keystr;
#line 337
        char *  value;
#line 337
        long  replace;
#line 337
        long  null;
#line 337
        long  mode;
#line 337
        
#line 337
get_string_arg(env, 1, &dbname);
#line 337
        get_string_arg(env, 2, &keystr);
#line 337
        get_string_arg(env, 3, &value);
#line 337
        get_numeric_arg(env, 4, &replace);
#line 337
        get_numeric_arg(env, 5, &null);
#line 337
        get_numeric_arg(env, 6, &mode);
#line 337
        
#line 337
get_numeric_arg(env, 0, &__bi_argcnt);
#line 337
        adjust_stack(env, __bi_argcnt + 1);
#line 337

#line 337

#line 337
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 337
		prog_trace(env, "dbinsert %s %s %s %lu %lu %lu",dbname, keystr, value, ((__bi_argcnt > 3) ? replace : 0), ((__bi_argcnt > 4) ? null : 0), ((__bi_argcnt > 5) ? mode : 0));;

{
	int rc;
	mu_dbm_file_t db;
	struct mu_dbm_datum key;
	struct mu_dbm_datum contents;
	const struct db_prop *prop = db_prop_lookup(dbname);
	const char *errstr;
	
	rc = _open_dbm(&db, dbname, MU_STREAM_RDWR,
		       ((__bi_argcnt > 5) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
#line 350
		       prop ? prop->hint : NULL);
	if (rc)
		(
#line 352
	env_throw_bi(env, mfe_dbfailure, "dbinsert", _("mf_dbm_open(%s) failed: %s"),dbname,mu_strerror(rc))
#line 352
);
#line 356
	memset(&key, 0, sizeof key);
	key.mu_dptr = keystr;
	key.mu_dsize = strlen(keystr);
	if (((__bi_argcnt > 4) ? null : prop && prop->null))
		key.mu_dsize++;

	memset(&contents, 0, sizeof contents);
	contents.mu_dptr = value;
	contents.mu_dsize = strlen(value) + 1;
	
	rc = mu_dbm_store(db, &key, &contents, ((__bi_argcnt > 3) ? replace : 0));
	if (rc && rc != MU_ERR_EXISTS)
		errstr = mu_dbm_strerror(db);
	mu_dbm_destroy(&db);
	if (rc == MU_ERR_EXISTS)
		(
#line 371
	env_throw_bi(env, mfe_exists, "dbinsert", _("record already exists"))
#line 371
);
	
		if (!(rc == 0))
#line 373
		(
#line 373
	env_throw_bi(env, mfe_dbfailure, "dbinsert", _("failed to insert data to %s: %s %s: %s"),dbname,keystr,value,errstr)
#line 373
)
#line 379
;
}

#line 381
        env_function_cleanup_flush(env, NULL);
#line 381
	return;
#line 381
}

void
#line 383
bi_dbdel(eval_environ_t env)
#line 383

#line 383

#line 383 "db.bi"
{
#line 383
	
#line 383

#line 383
long __bi_argcnt;
#line 383
char *  dbname;
#line 383
        char *  keystr;
#line 383
        long  null;
#line 383
        long  mode;
#line 383
        
#line 383
get_string_arg(env, 1, &dbname);
#line 383
        get_string_arg(env, 2, &keystr);
#line 383
        get_numeric_arg(env, 3, &null);
#line 383
        get_numeric_arg(env, 4, &mode);
#line 383
        
#line 383
get_numeric_arg(env, 0, &__bi_argcnt);
#line 383
        adjust_stack(env, __bi_argcnt + 1);
#line 383

#line 383

#line 383
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 383
		prog_trace(env, "dbdel %s %s %lu %lu",dbname, keystr, ((__bi_argcnt > 2) ? null : 0), ((__bi_argcnt > 3) ? mode : 0));;

{
	mu_dbm_file_t db;
	struct mu_dbm_datum key;
	int rc;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	rc = _open_dbm(&db, dbname, MU_STREAM_RDWR,
		       ((__bi_argcnt > 3) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
#line 394
		       prop ? prop->hint : NULL);
		if (!(rc == 0))
#line 395
		(
#line 395
	env_throw_bi(env, mfe_dbfailure, "dbdel", _("mf_dbm_open(%s) failed: %s"),dbname,mu_strerror(rc))
#line 395
)
#line 399
;
	memset(&key, 0, sizeof key);
	key.mu_dptr = keystr;
	key.mu_dsize = strlen(keystr);
	if (((__bi_argcnt > 2) ? null : prop && prop->null))
		key.mu_dsize++;
	rc = mu_dbm_delete(db, &key);
	if (rc && rc != MU_ERR_FAILURE)
		mu_error(_("error deleting %s from %s: %s"),
			 keystr, dbname, mu_dbm_strerror(db));
	mu_dbm_destroy(&db);
		if (!(rc == 0))
#line 410
		(
#line 410
	env_throw_bi(env, mfe_dbfailure, "dbdel", _("failed to delete data `%s' from `%s': %s"),keystr,dbname,mu_strerror(rc))
#line 410
)
#line 415
;
}

#line 417
        env_function_cleanup_flush(env, NULL);
#line 417
	return;
#line 417
}


#define NUMDB 128

struct db_tab {
	int used;
	mu_dbm_file_t db;
	struct mu_dbm_datum key;
};

static void *
alloc_db_tab()
{
	return xcalloc(NUMDB, sizeof(struct db_tab));
}

static void
close_db_tab(struct db_tab *dbt)
{
	if (dbt->used) {
		mu_dbm_datum_free(&dbt->key);
		mu_dbm_destroy(&dbt->db);
		dbt->used = 0;
	}
}

static void
destroy_db_tab(void *data)
{
	int i;
	struct db_tab *db = data;
	for (i = 0; i < NUMDB; i++) 
		close_db_tab(db + i);
	free(db);
}


#line 454

#line 454
static int DBTAB_id;
#line 454 "db.bi"
;

static int
new_db_tab(struct db_tab *dbt)
{
	int i;
	for (i = 0; i < NUMDB; i++)
		if (!dbt[i].used) {
			dbt[i].used = 1;
			return i;
		}
	return -1;
}
			


void
#line 470
bi_dbfirst(eval_environ_t env)
#line 470

#line 470

#line 470 "db.bi"
{
#line 470
	
#line 470

#line 470

#line 470
char *  dbname;
#line 470
        
#line 470
get_string_arg(env, 0, &dbname);
#line 470
        
#line 470

#line 470
        adjust_stack(env, 1);
#line 470

#line 470

#line 470
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 470
		prog_trace(env, "dbfirst %s",dbname);;
#line 470

{
	int rc;
	int n;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id);
	mu_dbm_file_t db;
	struct mu_dbm_datum key;
	const struct db_prop *prop = db_prop_lookup(dbname);

	rc = _open_dbm(&db, dbname, MU_STREAM_READ,
		       prop ? prop->mode : DEFAULT_DB_MODE,
		       prop ? prop->hint : NULL);
		if (!(rc == 0))
#line 482
		(
#line 482
	env_throw_bi(env, mfe_dbfailure, "dbfirst", _("mf_dbm_open(%s) failed: %s"),dbname,mu_strerror(rc))
#line 482
)
#line 486
;
	memset(&key, 0, sizeof key);
	rc = mu_dbm_firstkey(db, &key);
	if (rc) {
		if (rc == MU_ERR_NOENT) {
			mu_dbm_destroy(&db);
			
#line 492
do {
#line 492
  push(env, (STKVAL) (0));
#line 492
  goto endlab;
#line 492
} while (0);
		} else if (rc) {
			mu_dbm_destroy(&db);
			(
#line 495
	env_throw_bi(env, mfe_dbfailure, "dbfirst", _("mf_dbm_firstkey failed: %s"),mu_strerror(rc))
#line 495
);
#line 498
		}
	}
	n = new_db_tab(dbt);
		if (!(n >= 0))
#line 501
		(
#line 501
	env_throw_bi(env, mfe_failure, "dbfirst", _("no more database entries available"))
#line 501
)
#line 503
;
	dbt += n;
	dbt->db = db;
	dbt->key = key;
	
#line 507
do {
#line 507
  push(env, (STKVAL) (n));
#line 507
  goto endlab;
#line 507
} while (0);
}
endlab:
#line 509
        env_function_cleanup_flush(env, NULL);
#line 509
	return;
#line 509
}

void
#line 511
bi_dbnext(eval_environ_t env)
#line 511

#line 511

#line 511 "db.bi"
{
#line 511
	
#line 511

#line 511

#line 511
long  dn;
#line 511
        
#line 511
get_numeric_arg(env, 0, &dn);
#line 511
        
#line 511

#line 511
        adjust_stack(env, 1);
#line 511

#line 511

#line 511
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 511
		prog_trace(env, "dbnext %lu",dn);;
#line 511

{
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	struct mu_dbm_datum nextkey;
	int rc;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 517
		(
#line 517
	env_throw_bi(env, mfe_range, "dbnext", _("invalid database descriptor"))
#line 517
)
#line 519
;

	memset (&nextkey, 0, sizeof nextkey);
	rc = mu_dbm_nextkey(dbt->db, &nextkey);
	if (rc) {
		if (rc == MU_ERR_FAILURE)
			mu_error(_("mu_dbm_nextkey: %s"),
				 mu_dbm_strerror(dbt->db));
		close_db_tab(dbt);
		
#line 528
do {
#line 528
  push(env, (STKVAL) (0));
#line 528
  goto endlab;
#line 528
} while (0);
	}
	mu_dbm_datum_free(&nextkey);
	dbt->key = nextkey;
	
#line 532
do {
#line 532
  push(env, (STKVAL) (1));
#line 532
  goto endlab;
#line 532
} while (0);
}
endlab:
#line 534
        env_function_cleanup_flush(env, NULL);
#line 534
	return;
#line 534
}

void
#line 536
bi_dbkey(eval_environ_t env)
#line 536

#line 536

#line 536 "db.bi"
{
#line 536
	
#line 536

#line 536

#line 536
long  dn;
#line 536
        
#line 536
get_numeric_arg(env, 0, &dn);
#line 536
        
#line 536

#line 536
        adjust_stack(env, 1);
#line 536

#line 536

#line 536
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 536
		prog_trace(env, "dbkey %lu",dn);;
#line 536

{
	size_t off, len;
	char *s;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 542
		(
#line 542
	env_throw_bi(env, mfe_range, "dbkey", _("invalid database descriptor"))
#line 542
)
#line 544
;
	
	len = dbt->key.mu_dsize;
	s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
	memcpy(s, dbt->key.mu_dptr, len);
	s[len] = 0;
	
#line 550
do {
#line 550
  push(env, (STKVAL) (off));
#line 550
  goto endlab;
#line 550
} while (0);
}
endlab:
#line 552
        env_function_cleanup_flush(env, NULL);
#line 552
	return;
#line 552
}

void
#line 554
bi_dbvalue(eval_environ_t env)
#line 554

#line 554

#line 554 "db.bi"
{
#line 554
	
#line 554

#line 554

#line 554
long  dn;
#line 554
        
#line 554
get_numeric_arg(env, 0, &dn);
#line 554
        
#line 554

#line 554
        adjust_stack(env, 1);
#line 554

#line 554

#line 554
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 554
		prog_trace(env, "dbvalue %lu",dn);;
#line 554

{
	int rc;
	size_t off, len;
	char *s;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	struct mu_dbm_datum contents;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 562
		(
#line 562
	env_throw_bi(env, mfe_range, "dbvalue", _("invalid database descriptor"))
#line 562
)
#line 564
;

	memset(&contents, 0, sizeof contents);
	rc = mu_dbm_fetch(dbt->db, &dbt->key, &contents);
		if (!(rc == 0))
#line 568
		(
#line 568
	env_throw_bi(env, mfe_dbfailure, "dbvalue", _("cannot fetch key: %s"),rc == MU_ERR_FAILURE ?
		   mu_dbm_strerror(dbt->db) : mu_strerror (rc))
#line 568
)
#line 572
;
	
	len = contents.mu_dsize;
	s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
	memcpy(s, contents.mu_dptr, len);
	s[len] = 0;
	mu_dbm_datum_free(&contents);
	
#line 579
do {
#line 579
  push(env, (STKVAL) (off));
#line 579
  goto endlab;
#line 579
} while (0);
}
endlab:
#line 581
        env_function_cleanup_flush(env, NULL);
#line 581
	return;
#line 581
}


enum greylist_semantics
{
	greylist_traditional,
	greylist_ct
};

static enum greylist_semantics greylist_semantics = greylist_traditional;

/* #pragma greylist {traditional|gray|ct|con-tassios}*/
#line 593 "db.bi"

#line 593
static void _pragma_greylist (int argc, char **argv, const char *text)
#line 593

{
	if (strcmp(argv[1], "traditional") == 0
	    || strcmp(argv[1], "gray") == 0)
		greylist_semantics = greylist_traditional;
	else if (strcmp(argv[1], "ct") == 0
		 || strcmp(argv[1], "con-tassios") == 0)
		greylist_semantics = greylist_ct;
	else
		/* TRANSLATORS: Do not translate keywords:
		   traditional, gray, ct, con-tassios */
		parse_error(_("unknown semantics; allowed values are: "
			      "traditional (or gray) and "
			      "ct (or con-tassios)"));
}

/* FIXME: Duplicated in lib/cache.c */
static char *
format_timestr(time_t timestamp, char *timebuf, size_t bufsize)
{
	struct tm tm;
	gmtime_r(&timestamp, &tm);
	strftime(timebuf, bufsize, "%c", &tm);
	return timebuf;
}

static size_t greylist_seconds_left_loc
#line 619 "db.bi"
;

/* The traditional (aka gray's) greylist implementation: the greylist
   database keeps the time the greylisting was activated.
 */
static int
do_greylist_traditional(eval_environ_t env, char *email, long interval)
{
	int rc;
	mu_dbm_file_t db;
	struct mu_dbm_datum key;
	struct mu_dbm_datum contents;
	int readonly = 0;
	time_t now;

	rc = _open_dbm(&db, greylist_format->dbname, MU_STREAM_RDWR, 0600,
		       NULL);
	if (rc) {
		rc = _open_dbm(&db, greylist_format->dbname,
			       MU_STREAM_READ, 0600, NULL);
		readonly = 1;
	}
        	if (!(rc == 0))
#line 641
		(
#line 641
	env_throw_bi(env, mfe_dbfailure, NULL, _("mf_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_strerror(rc))
#line 641
)
;
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	key.mu_dptr = email;
	key.mu_dsize = strlen(email)+1;

	time(&now);
	rc = mu_dbm_fetch(db, &key, &contents);
	if (rc == 0) {
		time_t timestamp, diff;

			if (!(contents.mu_dsize == sizeof timestamp))
#line 654
		(
#line 654
	env_throw_bi(env, mfe_dbfailure, NULL, _("greylist database %s has wrong data size"),greylist_format->dbname)
#line 654
)
#line 657
;
		
		timestamp = *(time_t*) contents.mu_dptr;
		diff = now - timestamp;

		if (mu_debug_level_p(debug_handle, MU_DEBUG_TRACE5)) {
			char timebuf[32];
			mu_debug_log("%s entered greylist database on %s, "
				     "%ld seconds ago",
				     email,
				     format_timestr(timestamp, timebuf,
						    sizeof timebuf),
				     (long) diff);
		}
		
		if (diff < interval) {
			diff = interval - diff;
			
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (diff);
			
			
#line 677

#line 677
mu_debug(debug_handle, MU_DEBUG_TRACE6,("%s still greylisted (for %lu sec.)",
			          email,
			          (unsigned long) diff));
#line 681
			rc = 1;
		} else if (diff > greylist_format->expire_interval) {
			
#line 683

#line 683
mu_debug(debug_handle, MU_DEBUG_TRACE6,("greylist record for %s expired", email));
#line 685
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
			if (!readonly) {
				memcpy(contents.mu_dptr, &now, sizeof now);
				rc = mu_dbm_store(db, &key, &contents, 1);
				if (rc)
					mu_error(_("cannot insert datum `%-.*s' in "
						   "greylist database %s: %s"),
			                         key.mu_dsize,
						 (char*)key.mu_dptr,
						 greylist_format->dbname,
						 rc == MU_ERR_FAILURE ?
						  mu_dbm_strerror(db) :
						  mu_strerror(rc));
			} else
				
#line 699

#line 699
mu_debug(debug_handle, MU_DEBUG_TRACE6,("database opened in readonly mode: "
			  	          "not updating"));
#line 702
			rc = 1;
		} else {
			
#line 704

#line 704
mu_debug(debug_handle, MU_DEBUG_TRACE6,("%s finished greylisting period", email));
#line 706
			rc = 0;
		}
		mu_dbm_datum_free(&contents);
	} else if (!readonly) {
		
#line 710

#line 710
mu_debug(debug_handle, MU_DEBUG_TRACE6,("greylisting %s", email));
		*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
		contents.mu_dptr = (void*)&now;
		contents.mu_dsize = sizeof now;
		rc = mu_dbm_store(db, &key, &contents, 1);
		if (rc)
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 key.mu_dsize, (char*)key.mu_dptr,
				 greylist_format->dbname,
				 rc == MU_ERR_FAILURE ?
				  mu_dbm_strerror(db) : mu_strerror(rc));
		rc = 1;
	} else
		rc = 0;

	mu_dbm_destroy(&db);

	return rc;
}

/* Implementation of the is_greylisted predicate has no sense for 
   traditional greylist databases, because greylisting interval is
   not known beforehand.
   FIXME: keep the reference below up to date.
 */
static int
is_greylisted_traditional(eval_environ_t env, char *email)
{
	(
#line 739
	env_throw_bi(env, mfe_failure, NULL, _("is_greylisted is not implemented for traditional greylist databases; "
		   "see documentation, chapter %s %s for more info"),"4.12.1.23","Greylisting functions")
#line 739
);
#line 743
	return 0;
}

/* New greylist implementation (by Con Tassios): the database keeps
   the time the greylisting period is set to expire (`interval' seconds
   from now)
*/
static int
do_greylist_ct(eval_environ_t env, char *email, long interval)
{
	int rc;
	mu_dbm_file_t db;
	struct mu_dbm_datum key;
	struct mu_dbm_datum contents;
	int readonly = 0;
	time_t now;

	rc = _open_dbm(&db, greylist_format->dbname, MU_STREAM_RDWR, 0600,
		       NULL);
	if (rc) {
		rc = _open_dbm(&db, greylist_format->dbname,
			       MU_STREAM_READ, 0600, NULL);
		readonly = 1;
	}
        	if (!(rc == 0))
#line 767
		(
#line 767
	env_throw_bi(env, mfe_dbfailure, NULL, _("mf_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_strerror(rc))
#line 767
)
;
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	key.mu_dptr = email;
	key.mu_dsize = strlen(email) + 1;

	time(&now);
	rc = mu_dbm_fetch(db, &key, &contents);
	if (rc == 0) {
		time_t timestamp;

			if (!(contents.mu_dsize == sizeof timestamp))
#line 780
		(
#line 780
	env_throw_bi(env, mfe_dbfailure, NULL, _("greylist database %s has wrong data size"),greylist_format->dbname)
#line 780
)
#line 783
;
		
		timestamp = *(time_t*) contents.mu_dptr;

		if (now < timestamp) {
			time_t diff = timestamp - now;
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (diff);
			
			
#line 791

#line 791
mu_debug(debug_handle, MU_DEBUG_TRACE6,("%s still greylisted (for %lu sec.)",
			          email,
                                  (unsigned long) diff));
#line 795
			rc = 1;
		} else if (now - timestamp >
			   greylist_format->expire_interval) {
			
#line 798

#line 798
mu_debug(debug_handle, MU_DEBUG_TRACE6,("greylist record for %s expired", email));
#line 800
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
			if (!readonly) {
				now += interval;
				memcpy(contents.mu_dptr, &now, sizeof now);
				rc = mu_dbm_store(db, &key, &contents, 1);
				if (rc)
					mu_error(_("Cannot insert datum "
						   "`%-.*s' in greylist "
						   "database %s: %s"),
			                         key.mu_dsize,
						 (char*)key.mu_dptr,
						 greylist_format->dbname,
						 rc == MU_ERR_FAILURE ?
						   mu_dbm_strerror(db) :
						   mu_strerror(rc));
			} else
				
#line 816

#line 816
mu_debug(debug_handle, MU_DEBUG_TRACE6,("database opened in readonly mode: "
				          "not updating"));
#line 819
			rc = 1;
		} else {
			
#line 821

#line 821
mu_debug(debug_handle, MU_DEBUG_TRACE6,("%s finished greylisting period", email));
#line 823
			rc = 0;
		}
		mu_dbm_datum_free(&contents);
	} else if (!readonly) {
		
#line 827

#line 827
mu_debug(debug_handle, MU_DEBUG_TRACE6,("greylisting %s", email));
		*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
		now += interval;
		contents.mu_dptr = (void*)&now;
		contents.mu_dsize = sizeof now;
		rc = mu_dbm_store(db, &key, &contents, 1);
		if (rc)
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 key.mu_dsize, (char*)key.mu_dptr,
				 greylist_format->dbname,
				 rc == MU_ERR_FAILURE ?
				  mu_dbm_strerror(db) : mu_strerror(rc));
		rc = 1;
	} else
		rc = 0;
	
	mu_dbm_destroy(&db);
	
	return rc;
}

/* The `is_greylisted' predicate for new databases */
static int
is_greylisted_ct(eval_environ_t env, char *email)
{
	int rc;
	mu_dbm_file_t db;
	struct mu_dbm_datum key;
	struct mu_dbm_datum contents;
	int readonly = 0;
	time_t now;

	rc = _open_dbm(&db, greylist_format->dbname, MU_STREAM_RDWR, 0600,
		       NULL);
	if (rc) {
		rc = _open_dbm(&db, greylist_format->dbname,
			       MU_STREAM_READ, 0600, NULL);
		readonly = 1;
	}
		if (!(rc == 0))
#line 867
		(
#line 867
	env_throw_bi(env, mfe_dbfailure, NULL, _("mf_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_strerror(rc))
#line 867
)
;

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	key.mu_dptr = email;
	key.mu_dsize = strlen(email) + 1;

	time(&now);
	rc = mu_dbm_fetch(db, &key, &contents);
	if (rc == 0) {
		time_t timestamp;

			if (!(contents.mu_dsize == sizeof timestamp))
#line 880
		(
#line 880
	env_throw_bi(env, mfe_dbfailure, NULL, _("greylist database %s has wrong data size"),greylist_format->dbname)
#line 880
)
#line 883
;
                
		timestamp = *(time_t*) contents.mu_dptr;
                
		rc = timestamp > now;
                if (rc)
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (timestamp - now);

		mu_dbm_datum_free(&contents);
	} else  
		rc = 0;

	mu_dbm_destroy(&db);

	return rc;
}

struct greylist_class {
	int (*gl_fun)(eval_environ_t, char *, long);
	int (*gl_pred)(eval_environ_t, char *);
};

struct greylist_class greylist_class[] = {
	{ do_greylist_traditional, is_greylisted_traditional },
	{ do_greylist_ct, is_greylisted_ct }
};

/* greylist(key, interval)

   Returns true if the key is greylisted, false if it's OK to
   deliver mail.
 */
void
#line 915
bi_greylist(eval_environ_t env)
#line 915

#line 915

#line 915 "db.bi"
{
#line 915
	
#line 915

#line 915

#line 915
char *  email;
#line 915
        long  interval;
#line 915
        
#line 915
get_string_arg(env, 0, &email);
#line 915
        get_numeric_arg(env, 1, &interval);
#line 915
        
#line 915

#line 915
        adjust_stack(env, 2);
#line 915

#line 915

#line 915
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 915
		prog_trace(env, "greylist %s %lu",email, interval);;
#line 915

{
	
#line 917
do {
#line 917
  push(env, (STKVAL) (greylist_class[greylist_semantics].gl_fun(env, email,
#line 917
							    interval)));
#line 917
  goto endlab;
#line 917
} while (0);
#line 919
}
endlab:
#line 920
        env_function_cleanup_flush(env, NULL);
#line 920
	return;
#line 920
}

/* is_greylisted(key)
   
   Returns true if the key is greylisted, otherwise false
   
*/
void
#line 927
bi_is_greylisted(eval_environ_t env)
#line 927

#line 927

#line 927 "db.bi"
{
#line 927
	
#line 927

#line 927

#line 927
char *  email;
#line 927
        
#line 927
get_string_arg(env, 0, &email);
#line 927
        
#line 927

#line 927
        adjust_stack(env, 1);
#line 927

#line 927

#line 927
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 927
		prog_trace(env, "is_greylisted %s",email);;
#line 927

{
	
#line 929
do {
#line 929
  push(env, (STKVAL) (greylist_class[greylist_semantics].gl_pred(env, email)));
#line 929
  goto endlab;
#line 929
} while (0);
}
endlab:
#line 931
        env_function_cleanup_flush(env, NULL);
#line 931
	return;
#line 931
}

void
#line 933
bi_db_name(eval_environ_t env)
#line 933

#line 933

#line 933 "db.bi"
{
#line 933
	
#line 933

#line 933

#line 933
char * MFL_DATASEG fmtid;
#line 933
        
#line 933
get_string_arg(env, 0, &fmtid);
#line 933
        
#line 933

#line 933
        adjust_stack(env, 1);
#line 933

#line 933

#line 933
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 933
		prog_trace(env, "db_name %s",fmtid);;
#line 933

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 936
		(
#line 936
	env_throw_bi(env, mfe_not_found, "db_name", _("no such db format: %s"),fmtid)
#line 936
)
#line 938
;
	do {pushs(env, fmt->dbname);
#line 939
goto endlab;
#line 939
} while (0);
}
endlab:
#line 941
        env_function_cleanup_flush(env, NULL);
#line 941
	return;
#line 941
}

void
#line 943
bi_db_get_active(eval_environ_t env)
#line 943

#line 943

#line 943 "db.bi"
{
#line 943
	
#line 943

#line 943

#line 943
char *  fmtid;
#line 943
        
#line 943
get_string_arg(env, 0, &fmtid);
#line 943
        
#line 943

#line 943
        adjust_stack(env, 1);
#line 943

#line 943

#line 943
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 943
		prog_trace(env, "db_get_active %s",fmtid);;
#line 943

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 946
		(
#line 946
	env_throw_bi(env, mfe_not_found, "db_get_active", _("no such db format: %s"),fmtid)
#line 946
)
#line 948
;
	
#line 949
do {
#line 949
  push(env, (STKVAL) (fmt->enabled));
#line 949
  goto endlab;
#line 949
} while (0);
}
endlab:
#line 951
        env_function_cleanup_flush(env, NULL);
#line 951
	return;
#line 951
}

void
#line 953
bi_db_set_active(eval_environ_t env)
#line 953

#line 953

#line 953 "db.bi"
{
#line 953
	
#line 953

#line 953

#line 953
char *  fmtid;
#line 953
        long  active;
#line 953
        
#line 953
get_string_arg(env, 0, &fmtid);
#line 953
        get_numeric_arg(env, 1, &active);
#line 953
        
#line 953

#line 953
        adjust_stack(env, 2);
#line 953

#line 953

#line 953
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 953
		prog_trace(env, "db_set_active %s %lu",fmtid, active);;
#line 953

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 956
		(
#line 956
	env_throw_bi(env, mfe_not_found, "db_set_active", _("no such db format: %s"),fmtid)
#line 956
)
#line 958
;
	fmt->enabled = active;
}

#line 961
        env_function_cleanup_flush(env, NULL);
#line 961
	return;
#line 961
}

void
#line 963
bi_db_expire_interval(eval_environ_t env)
#line 963

#line 963

#line 963 "db.bi"
{
#line 963
	
#line 963

#line 963

#line 963
char *  fmtid;
#line 963
        
#line 963
get_string_arg(env, 0, &fmtid);
#line 963
        
#line 963

#line 963
        adjust_stack(env, 1);
#line 963

#line 963

#line 963
	if (builtin_module_trace(BUILTIN_IDX_db))
#line 963
		prog_trace(env, "db_expire_interval %s",fmtid);;
#line 963

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 966
		(
#line 966
	env_throw_bi(env, mfe_not_found, "db_expire_interval", _("no such db format: %s"),fmtid)
#line 966
)
#line 968
;
	
#line 969
do {
#line 969
  push(env, (STKVAL) (fmt->expire_interval));
#line 969
  goto endlab;
#line 969
} while (0);
}
endlab:
#line 971
        env_function_cleanup_flush(env, NULL);
#line 971
	return;
#line 971
}


#line 973
void
#line 973
db_init_builtin()
#line 973
{
#line 973
		debug_handle = mu_debug_register_category("bi_db");
#line 973

#line 973
	#line 81 "db.bi"

#line 81
install_pragma("dbprop", 3, 5, _pragma_dbprop);
#line 271 "db.bi"
va_builtin_install_ex("dbmap", bi_dbmap, 0, dtype_number, 3, 1, 0|0, dtype_string, dtype_string, dtype_number);
#line 282 "db.bi"
va_builtin_install_ex("dbget", bi_dbget, 0, dtype_string, 4, 2, 0|0, dtype_string, dtype_string, dtype_string, dtype_number);
#line 294 "db.bi"
va_builtin_install_ex("dbput", bi_dbput, 0, dtype_unspecified, 5, 2, 0|0, dtype_string, dtype_string, dtype_string, dtype_number, dtype_number);
#line 337 "db.bi"
va_builtin_install_ex("dbinsert", bi_dbinsert, 0, dtype_unspecified, 6, 3, 0|0, dtype_string, dtype_string, dtype_string, dtype_number, dtype_number, dtype_number);
#line 383 "db.bi"
va_builtin_install_ex("dbdel", bi_dbdel, 0, dtype_unspecified, 4, 2, 0|0, dtype_string, dtype_string, dtype_number, dtype_number);
#line 454 "db.bi"
DBTAB_id = builtin_priv_register(alloc_db_tab, destroy_db_tab,
#line 454
NULL);
#line 470 "db.bi"
va_builtin_install_ex("dbfirst", bi_dbfirst, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 511 "db.bi"
va_builtin_install_ex("dbnext", bi_dbnext, 0, dtype_number, 1, 0, 0|0, dtype_number);
#line 536 "db.bi"
va_builtin_install_ex("dbkey", bi_dbkey, 0, dtype_string, 1, 0, 0|0, dtype_number);
#line 554 "db.bi"
va_builtin_install_ex("dbvalue", bi_dbvalue, 0, dtype_string, 1, 0, 0|0, dtype_number);
#line 593 "db.bi"

#line 593
install_pragma("greylist", 2, 2, _pragma_greylist);
#line 619 "db.bi"
	builtin_variable_install("greylist_seconds_left", dtype_number, SYM_VOLATILE, &greylist_seconds_left_loc);
#line 915 "db.bi"
va_builtin_install_ex("greylist", bi_greylist, 0, dtype_number, 2, 0, 0|0, dtype_string, dtype_number);
#line 927 "db.bi"
va_builtin_install_ex("is_greylisted", bi_is_greylisted, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 933 "db.bi"
va_builtin_install_ex("db_name", bi_db_name, 0, dtype_string, 1, 0, 0|0, dtype_string);
#line 943 "db.bi"
va_builtin_install_ex("db_get_active", bi_db_get_active, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 953 "db.bi"
va_builtin_install_ex("db_set_active", bi_db_set_active, 0, dtype_unspecified, 2, 0, 0|0, dtype_string, dtype_number);
#line 963 "db.bi"
va_builtin_install_ex("db_expire_interval", bi_db_expire_interval, 0, dtype_number, 1, 0, 0|0, dtype_string);

#line 973 "db.bi"
	
#line 973
}
#line 973 "db.bi"

#line 973


#line 763 "../../src/builtin/snarf.m4"

#line 763

